<?php

namespace MapTimePlugin;

use FML\Controls\Frame;
use FML\Controls\Labels\Label_Button;
use FML\Controls\Labels\Label_Text;
use FML\Controls\Quad;
use FML\Controls\Quads\Quad_BgsPlayerCard;
use FML\Controls\Quads\Quad_Icons128x128_1;
use FML\Controls\Quads\Quad_Icons64x64_1;
use FML\ManiaLink;
use FML\Script\Features\Paging;
use FML\Script\Script;
use ManiaControl\Admin\AuthenticationManager;
use ManiaControl\Commands\CommandListener;
use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Logger;
use ManiaControl\ManiaControl;
use ManiaControl\Manialinks\IconManager;
use ManiaControl\Manialinks\LabelLine;
use ManiaControl\Manialinks\ManialinkManager;
use ManiaControl\Manialinks\ManialinkPageAnswerListener;
use ManiaControl\Maps\Map;
use ManiaControl\Maps\MapManager;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Plugins\Plugin;
use ManiaControl\Utils\Formatter;
use Maniaplanet\DedicatedServer\InvalidArgumentException;

/**
 * ManiaControl MapTime-Plugin for managing individual map/track play times
 *
 * @author
 * @copyright
 * @version   0.2
 * @license   http://www.gnu.org/licenses/ GNU General Public License, Version 3
 */
class MapTimePlugin implements CommandListener, Plugin, CallbackListener, ManialinkPageAnswerListener {
	/*
	 * Constants
	 */
	const PLUGIN_ID           = 142;
	const PLUGIN_VERSION      = 0.2;
	const PLUGIN_NAME         = 'MapTimePlugin';
	const PLUGIN_AUTHOR       = 'crimplene';
	const TABLE_MAP_PLAY_TIME = 'mc_maptime';

	const SETTING_PERMISSION_MAP_TIME_PLUGIN    = 'MapTime-Plugin';

	const SETTING_TEST_WIDGET_ACTIVATED = 'Test-Widget Activated';
	const SETTING_TEST_WIDGET_POSX      = 'Test-Widget-Position: X';
	const SETTING_TEST_WIDGET_POSY      = 'Test-Widget-Position: Y';
	const SETTING_TEST_WIDGET_WIDTH     = 'Test-Widget-Size: Width';
	const SETTING_TEST_WIDGET_HEIGHT    = 'Test-Widget-Size: Height';

	const ACTION_LINE_CLICK             = 'MapTimePlugin.LineClick';

	private $numberOfDeleteClicks = array();

	/*
	 * Private properties
	 */
	/** @var ManiaControl $maniaControl */
	private $maniaControl = null;

	private $currentTimeLimit = null;

	public $frame = null;

	/**
	 * @see \ManiaControl\Plugins\Plugin::prepare()
	 * @param \ManiaControl\ManiaControl $maniaControl
	 */
	public static function prepare(ManiaControl $maniaControl) {
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getId()
	 */
	public static function getId() {
		return self::PLUGIN_ID;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getName()
	 */
	public static function getName() {
		return self::PLUGIN_NAME;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getVersion()
	 */
	public static function getVersion() {
		return self::PLUGIN_VERSION;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getAuthor()
	 */
	public static function getAuthor() {
		return self::PLUGIN_AUTHOR;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::getDescription()
	 */
	public static function getDescription() {
		return "This Plugin sets individual time limits for maps";
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::load()
	 * @param \ManiaControl\ManiaControl $maniaControl
	 * @return bool
	 */
	public function load(ManiaControl $maniaControl) {

		$this->maniaControl = $maniaControl;
		$this->initTables();

		/**
		 * Settings
		 */
		//$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_HIER_KOENNTEN_EINSTELLUNGEN_SEIN, false);

		$this->maniaControl->getAuthenticationManager()->definePermissionLevel(self::SETTING_PERMISSION_MAP_TIME_PLUGIN, AuthenticationManager::AUTH_LEVEL_SUPERADMIN);
		/**
		 * Callbacks
		 */
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_BEGINMAP, $this, 'handleBeginMap');
		$this->maniaControl->getCallbackManager()->registerCallbackListener(CallbackManager::CB_MP_PLAYERMANIALINKPAGEANSWER, $this, 'handleManialinkPageAnswer');

		$this->maniaControl->getCommandManager()->registerCommandListener(array('spt', 'setplaytime'), $this, 'chat_setPlayTime', true, 'Sets the time limit for the current map permanently');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('gpt', 'getplaytime'), $this, 'chat_getPlayTime', true, 'Displays the individual time limit for the current map');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('sst', 'setstandardtime'), $this, 'chat_setStandardTime', true, 'Sets the standard time limit');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('gst', 'getstandardtime'), $this, 'chat_getStandardTime', true, 'Displays the standard time limit');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('stts', 'setTimeToStandard'), $this, 'chat_setTimeToStandard', true, 'Resets the time limit of the current map to standard');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('mtpadd', 'timeadd', 'tadd'), $this, 'chat_addPlayTime', true, 'Adds minutes to the TimeLimit of the current map');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('mtpsubst', 'timesubst', 'tsubst'), $this, 'chat_substractPlayTime', true, 'Substracts minutes from the TimeLimit of the current map');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('mtpshow', 'timeshow', 'tshow'), $this, 'showWidget', true, 'Shows an Info Widget');
		$this->maniaControl->getCommandManager()->registerCommandListener(array('mtphelp', 'timehelp', 'thelp'), $this, 'chat_help', true, 'Shows all MapTimePlugin commands');

		$scriptSettings = $this->maniaControl->getClient()->getModeScriptSettings();
		$this->currentTimeLimit = $scriptSettings['S_TimeLimit'];

		$this->frame = new Frame();

		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEST_WIDGET_ACTIVATED, true);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEST_WIDGET_POSX, 0);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEST_WIDGET_POSY, 0);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEST_WIDGET_WIDTH, 100);
		$this->maniaControl->getSettingManager()->initSetting($this, self::SETTING_TEST_WIDGET_HEIGHT, 75);

		//TODO: React to Setting Change
		return true;
	}

	/**
	 * Initialize needed database tables
	 */
	private function initTables() {
		$mysqli = $this->maniaControl->getDatabase()->getMysqli();

		$query = "CREATE TABLE IF NOT EXISTS `" . self::TABLE_MAP_PLAY_TIME . "` (
				`uid` varchar(50) NOT NULL,
				`mapName` varchar(100) NOT NULL,
				`timelimit` int(11) NOT NULL,
				`timestamp` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
				PRIMARY KEY (`uid`)
				) ENGINE=MyISAM DEFAULT CHARSET=utf8 COLLATE=utf8_unicode_ci AUTO_INCREMENT=1;";
		$mysqli->query($query);

		$query = "SELECT * FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE uid = 'standard'";

		$result = $mysqli->query($query);

		if (!isset($result->num_rows) || $result->num_rows != 1) {
			$query = "INSERT INTO `" . self::TABLE_MAP_PLAY_TIME . "`
						(`uid`,
						`mapName`,
						`timelimit`)
						VALUES (
						'standard',
						'standard',
						30)";

			$mysqli->query($query);

			$this->maniaControl->getChat()->sendChat("Map Time Plugin initialized. Standard Time Limit set to 30 minutes.");
		}

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		return;
	}

	/**
	 * @see \ManiaControl\Plugins\Plugin::unload()
	 */
	public function unload() {
	}

	public function handleBeginMap() {

		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;

		$mapTime = $this->getMapTimeLimit($mapUID);

		if ( $mapTime && $mapTime >= 5 ) {
			try
			{
				$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $mapTime ) );
			}
			catch ( InvalidArgumentException $e )
			{
			}
			$this->currentTimeLimit = (int) $mapTime;
			$this->maniaControl->getChat()->sendChat("The time limit for this map is " . $mapTime . " minutes!");
		} else {
			$timeLimit = $this->getStandardTime();
			try
			{
				$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $timeLimit ) );
			}
			catch ( InvalidArgumentException $e )
			{
			}
			$this->currentTimeLimit = (int) $timeLimit;
			$this->maniaControl->getChat()->sendChat("Time Limit set to standard playing time of " . $timeLimit . " minutes!");
		}
	}

	/**
	 * @param array  $chat
	 * @param Player $admin
	 * @internal param int $minutes
	 */
	public function chat_setPlayTime(array $chat, Player $admin) {
		if (!$this->maniaControl->getAuthenticationManager()->checkPermission($admin, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$command = explode(" ", $chat[1][2]);
		$minutes = (int) $command[1];

		if (!is_int($minutes) || $minutes < 5) {

			$this->maniaControl->getChat()->sendChat("Something was wrong with the time. Either it was no integer or it is smaller than 5 minutes. No changes made");

			return;
		}

		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;
		$mapName   = Formatter::stripCodes($currentMap->rawName);

		$mysqli = $this->maniaControl->getDatabase()->getMysqli();

		$query = "SELECT * FROM " . self::TABLE_MAP_PLAY_TIME . " WHERE uid = '" . $mapUID . "'";

		$result = $mysqli->query($query);

		if (!isset($result->num_rows) || $result->num_rows != 1) {
			$query2 = "INSERT INTO " . self::TABLE_MAP_PLAY_TIME . "
						(uid,
						 mapName,
						 timelimit
						)
						VALUES
						('" . $mapUID . "',
						'" . $mapName . "',
						" . $minutes . "
						)";
		} else {
			$query2 = "UPDATE `" . self::TABLE_MAP_PLAY_TIME . "`
						SET timelimit = " . $minutes . "
						WHERE `uid`='" . $mapUID . "'";

		}

		//$query2 = mysqli_real_escape_string($mysqli, $query2);
		//$this->maniaControl->getChat()->sendChat($query2);

		//Logger::log("SQL-Befehl: ".$query2);

		$mysqli->query($query2);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		try
		{
			$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $minutes ) );
		}
		catch ( InvalidArgumentException $e )
		{
		}
		$this->currentTimeLimit = (int) $minutes;
		$this->maniaControl->getChat()->sendChat($admin->getAuthLevelName() . " " . $admin->login . " changed the time limit for this map to " . $minutes . " minutes!");

		return;
	}

	public function chat_getPlayTime(array $chat, Player $admin) {
		if (!AuthenticationManager::checkRight($admin, AuthenticationManager::AUTH_LEVEL_MODERATOR)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}
		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;

		$mapLimit = $this->getMapTimeLimit($mapUID);
		$standard = $this->getStandardTime();

		if( $mapLimit && $mapLimit >= 5 ) {
			$this->maniaControl->getChat()->sendChat("Time limit for this map is set to " . $mapLimit . " minutes.");
		} else {
			$this->maniaControl->getChat()->sendChat("There is no individual time limit set for this map. Standard time limit is ".$standard." minutes.");
		}
		return;
	}

	public function chat_setStandardTime(array $chat, Player $admin) {
		if (!$this->maniaControl->getAuthenticationManager()->checkPermission($admin, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$mysqli  = $this->maniaControl->getDatabase()->getMysqli();
		$command = explode(" ", $chat[1][2]);
		$minutes = (int) $command[1];

		if (!is_int($minutes) || $minutes < 5) {

			$this->maniaControl->getChat()->sendErrorToAdmins("Something was wrong with the standard time. Either it was no integer or it is smaller than 10 minutes. No changes made");

			return;
		}

		$query = "UPDATE `" . self::TABLE_MAP_PLAY_TIME . "`
						SET timelimit = " . $minutes . "
						WHERE `uid`='standard'";

		$mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		$this->maniaControl->getChat()->sendChat($admin->getAuthLevelName() . " " . $admin->login . " has set the standard time limit to " . $minutes . " minutes.");

		return;
	}

	public function chat_getStandardTime(array $chat, Player $admin) {
		if (!AuthenticationManager::checkRight($admin, AuthenticationManager::AUTH_LEVEL_MODERATOR)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$timeLimit = $this->getStandardTime();

		$this->maniaControl->getChat()->sendChat("Standard time limit is set to " . $timeLimit . " minutes.");

		return;
	}

	public function chat_setTimeToStandard(array $chat, Player $admin) {
		if (!$this->maniaControl->getAuthenticationManager()->checkPermission($admin, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$mysqli = $this->maniaControl->getDatabase()->getMysqli();

		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;

		$query = "DELETE FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE `uid` = '" . $mapUID . "'";

		$mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		$standardTime = $this->getStandardTime();

		try
		{
			$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $standardTime ) );
		}
		catch ( InvalidArgumentException $e )
		{
		}
		$this->currentTimeLimit = (int) $standardTime;

		$this->maniaControl->getChat()->sendChat($admin->getAuthLevelName() . " " . $admin->login . " has reset the time limit for this map to standard.");

		return;
	}

	public function chat_addPlayTime(array $chat, Player $admin) {
		if (!$this->maniaControl->getAuthenticationManager()->checkPermission($admin, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$command = explode(" ", $chat[1][2]);
		$minutesToAdd = (int) $command[1];
		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;
		$minutes = $this->getMapTimeLimit( $mapUID );

		if (!is_int($minutesToAdd)) {
			$this->maniaControl->getChat()->sendChat("Something was wrong with the time. It probably was no integers. No changes made.");
			return;
		}

		if( !isset($minutes) || $minutes < 5 ) {
			$minutes = $this->getStandardTime() + $minutesToAdd;
		} else {
			$minutes += $minutesToAdd;
		}

		if($minutes < 5){
			$this->maniaControl->getChat()->sendChat("Time Limit can't be under 5 minutes. No changes made.");
			return;
		}

		$mapName   = Formatter::stripCodes($currentMap->rawName);

		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query = "SELECT * FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE uid = '" . $mapUID . "'";
		$result = $mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		if (!isset($result->num_rows) || $result->num_rows != 1) {
			$query2 = "INSERT INTO `" . self::TABLE_MAP_PLAY_TIME . "`
						(`uid`,
						 `mapName`,
						 `timelimit`
						)
						VALUES
						('" . $mapUID . "',
						'" . $mapName . "',
						" . $minutes . "
						)";
		} else {
			$query2 = "UPDATE `" . self::TABLE_MAP_PLAY_TIME . "`
						SET timelimit = " . $minutes . "
						WHERE `uid`='" . $mapUID . "'";
		}

		$mysqli->query($query2);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		//TODO: da prüfen, speichern, melden und currentTimeLimit ändern oft zusammen vorkommen, können vermutlich alle Funktionen verschlankt werden, indem das in eine eigene kommt

		try
		{
			$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $minutes ) );
		}
		catch ( InvalidArgumentException $e )
		{
		}
		$this->currentTimeLimit = (int) $minutes;
		$this->maniaControl->getChat()->sendChat($admin->getAuthLevelName() . " " . $admin->login . " changed the time limit for this map to " . $minutes . " minutes!");

		return;
	}

	public function chat_substractPlayTime(array $chat, Player $admin) {
		if (!$this->maniaControl->getAuthenticationManager()->checkPermission($admin, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($admin);
			return;
		}

		$command = explode(" ", $chat[1][2]);
		$minutesToSubstract = (int) $command[1];
		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();
		$mapUID    = $currentMap->uid;
		$minutes = $this->getMapTimeLimit( $mapUID );
		$standardTime = $this->getStandardTime();

		// if no integer is given
		if (!is_int($minutesToSubstract) ) {
			$this->maniaControl->getChat()->sendChat("Something was wrong with the time. Either it was no integer or it becomes smaller than 5 minutes. No changes made");
			return;
		}

		// if a time limit is NOT already set
		if( !isset($minutes) || $minutes < 5 ){
			// if standard time minus entered time is to low
			if (($standardTime - $minutesToSubstract) < 5 ) {
				$this->maniaControl->getChat()->sendChat("Something was wrong with the time. Either it was no integer or it becomes smaller than 5 minutes. No changes made");
				return;
			} else {
				$minutes = $standardTime - $minutesToSubstract;
			}
		// if a time limit IS already set
		} else {
			// if new time limit is too low
			if (($minutes - $minutesToSubstract) < 5 ) {
				$this->maniaControl->getChat()->sendChat("Something was wrong with the time. Either it was no integer or it becomes smaller than 5 minutes. No changes made");
				return;
			} else {
				$minutes -= $minutesToSubstract;
			}
		}

		$mapName   = Formatter::stripCodes($currentMap->rawName);

		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query = "SELECT * FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE uid = '" . $mapUID . "'";
		$result = $mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		if (!isset($result->num_rows) || $result->num_rows != 1) {
			$query2 = "INSERT INTO `" . self::TABLE_MAP_PLAY_TIME . "`
						(`uid`,
						 `mapName`,
						 `timelimit`
						)
						VALUES
						('" . $mapUID . "',
						'" . $mapName . "',
						" . $minutes . "
						)";
		} else {
			$query2 = "UPDATE `" . self::TABLE_MAP_PLAY_TIME . "`
						SET timelimit = " . $minutes . "
						WHERE `uid`='" . $mapUID . "'";
		}

		$mysqli->query($query2);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		try
		{
			$this->maniaControl->getClient()->setModeScriptSettings( Array( 'S_TimeLimit' => (int) $minutes ) );
		}
		catch ( InvalidArgumentException $e )
		{
		}
		$this->currentTimeLimit = (int) $minutes;
		$this->maniaControl->getChat()->sendChat($admin->getAuthLevelName() . " " . $admin->login . " changed the time limit for this map to " . $minutes . " minutes!");

		return;
	}

	public function chat_help(array $chat, Player $player){

		$this->showHelpWidget($chat, $player);
		return;
	}

	public function showWidget(array $chat, Player $player) {
		if (!AuthenticationManager::checkRight($player, AuthenticationManager::AUTH_LEVEL_MODERATOR)) {
			$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
			return;
		}

		$playerLogin = $player->login;

		$width  = $this->maniaControl->getManialinkManager()->getStyleManager()->getListWidgetsWidth();
		$height = $this->maniaControl->getManialinkManager()->getStyleManager()->getListWidgetsHeight();

		// get List of TimeLimits
		$maps = $this->getTimeLimits();

		$currentMap = $this->maniaControl->getMapManager()->getCurrentMap();

		// create manialink
		$maniaLink = new ManiaLink(ManialinkManager::MAIN_MLID);
		$script    = $maniaLink->getScript();
		$paging    = new Paging();
		$script->addFeature($paging);

		// Main frame
		$frame = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultListFrame($script, $paging);
		$maniaLink->addChild($frame);

		// Start offsets
		$posX = -$width / 2;
		$posY = $height / 2;

		// Predefine Description Label
		$descriptionLabel = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultDescriptionLabel();
		$frame->addChild($descriptionLabel);

		$column1 = 5;
		$column2 = (int)($width/10);
		$column3 = (int)($width/4);
		$column4 = (int)($width/1.4);
		$column5 = (int)($width/1.25);

		// Headline
		$headFrame = new Frame();
		$frame->addChild($headFrame);
		$headFrame->setY($posY - 5);
		$labelLine = new LabelLine($headFrame);
		$labelLine->addLabelEntryText('ID', $posX +$column1);
		$labelLine->addLabelEntryText('MX ID', $posX + $column2);
		$labelLine->addLabelEntryText('Map Name', $posX + $column3);
		$labelLine->addLabelEntryText('TimeLimit', $posX + $column4);
		$labelLine->render();

		$index     = 0;
		$posY      = $height / 2 - 10;
		$pageFrame = null;

		$linesPerPage = (int)(((15/80)*$height)*1.2) - 1 ;
		$linesPerPage = $linesPerPage < 100 ? $linesPerPage - 1 : $linesPerPage;

		foreach ($maps as $map) {
			if ($index % $linesPerPage === 0) {
				$pageFrame = new Frame();
				$frame->addChild($pageFrame);
				$posY = $height / 2 - 10;
				$paging->addPageControl($pageFrame);
			}

			$mapFrame = new Frame();
			$pageFrame->addChild($mapFrame);

			if ($index % 2 === 0) {
				$lineQuad = new Quad_BgsPlayerCard();
				$mapFrame->addChild($lineQuad);
				$lineQuad->setSize($width, 4);
				$lineQuad->setSubStyle($lineQuad::SUBSTYLE_BgPlayerCardBig);
				$lineQuad->setZ(-0.001);
			}

			if ($map->id == $currentMap->index) {
				$current = "<--";
			} else {
				$current = "";
			}

//			$mxIcon      = $this->maniaControl->getManialinkManager()->getIconManager()->getIcon(IconManager::MX_ICON);
//			$mxIconHover = $this->maniaControl->getManialinkManager()->getIconManager()->getIcon(IconManager::MX_ICON_MOVER);
//
//			$mxQuad = new Quad();
//			$mapFrame->addChild($mxQuad);
//			$mxQuad->setSize(3, 3);
//			$mxQuad->setImageUrl($mxIcon);
//			$mxQuad->setImageFocusUrl($mxIconHover);
//			$mxQuad->setX($posX + 30);
//			$mxQuad->setAction(self::ACTION_LINE_CLICK.".".$map->uid.".".$playerLogin);
//			$mxQuad->addTooltipLabelFeature($descriptionLabel, "Delete individual Time Limit");
//			$mxQuad->setZ(0.01);

			if ($this->maniaControl->getAuthenticationManager()->checkPermission($player, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
				// remove map button
				$removeButton = new Label_Button();
				$mapFrame->addChild($removeButton);
				$removeButton->setX($posX + $column5);
				$removeButton->setZ(0.2);
				$removeButton->setSize(3, 3);
				$removeButton->setTextSize(1);
				$removeButton->setText('X');
				$removeButton->setTextColor('a00');

				$removeButton->addTooltipLabelFeature($descriptionLabel, "Delete individual Time Limit - Click 3 times!");

				$removeButton->setAction(self::ACTION_LINE_CLICK.".".$map->uid.".".$playerLogin);
			}

			$labelLine = new LabelLine($mapFrame);
			$labelLine->addLabelEntryText($map->id.$current, $posX + $column1, 13);
			$labelLine->addLabelEntryText($map->mxid, $posX + $column2, 52);
			$labelLine->addLabelEntryText($map->name.$current, $posX + $column3, 31);
			$labelLine->addLabelEntryText($map->time, $posX + $column4, $width / 2 - ($posX + 110));
			$labelLine->render();

			$mapFrame->setY($posY);

			$posY -= 4;
			$index++;
		}

		// Render and display xml
		$this->maniaControl->getManialinkManager()->displayWidget($maniaLink, $player, 'TimeLimits');
	}

	public function showHelpWidget(array $chat, Player $player) {

		$width  = $this->maniaControl->getManialinkManager()->getStyleManager()->getListWidgetsWidth();
		$height = $this->maniaControl->getManialinkManager()->getStyleManager()->getListWidgetsHeight();

		$maniaLink = new ManiaLink(ManialinkManager::MAIN_MLID);
		$script    = $maniaLink->getScript();

		// Main frame
		$frame = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultListFrame($script);
		$maniaLink->addChild($frame);

		// Start offsets
		$posX = -$width / 2;
		$posY = $height / 2;

		// Predefine Description Label
		$descriptionLabel = $this->maniaControl->getManialinkManager()->getStyleManager()->getDefaultDescriptionLabel();
		$frame->addChild($descriptionLabel);

		// Title
		$titleFrame = new Frame();
		$frame->addChild($titleFrame);
		$titleFrame->setY($posY - 5);

		$labelLine1 = new LabelLine($titleFrame);
		$labelLine1->addLabelEntryText('$w$oList of MapTimePlugin commands', $posX + 5);
		$labelLine1->render();

		// Headline
		$headFrame = new Frame();
		$frame->addChild($headFrame);
		$headFrame->setY($posY -10);

		$labelLine2 = new LabelLine($headFrame);
		$labelLine2->addLabelEntryText('$oAdmin Command', $posX + 5);
		$labelLine2->addLabelEntryText('$oExplanation', $posX + 45);
		$labelLine2->render();

		$index     = 0;
		$posY      = $height / 2 - 10;
		$pageFrame = null;

		$commands = array();

		$commands[] = array('command'=>'spt, setplaytime', 'explanation'=>'Sets an individual time limit for the current map permanently');
		$commands[] = array('command'=>'gpt, getplaytime', 'explanation'=>'Displays the individual time limit for the current map');
		$commands[] = array('command'=>'sst, setstandardtime', 'explanation'=>'Sets the standard time limit');
		$commands[] = array('command'=>'gst, getstandardtime', 'explanation'=>'Displays the standard time limit');
		$commands[] = array('command'=>'stts, setTimeToStandard', 'explanation'=>'Resets the time limit of the current map to standard');
		$commands[] = array('command'=>'mtpadd, timeadd, tadd', 'explanation'=>'Adds minutes to the time limit of the current map');
		$commands[] = array('command'=>'mtpsubst, timesubst, tsubst', 'explanation'=>'Substracts minutes from the time limit of the current map');
		$commands[] = array('command'=>'mtpshow, timeshow, tshow', 'explanation'=>'Displays an info widget with all maps that have an individual time limit');
		$commands[] = array('command'=>'mtphelp, timehelp, thelp', 'explanation'=>'Displays this help window');
		$posY -= 8;
		foreach ($commands as $command) {

			$pageFrame = new Frame();
			$frame->addChild($pageFrame);
			//$pageFrame->setY($posY -25);
			//$paging->addPageControl($pageFrame);

			$commandFrame = new Frame();
			$pageFrame->addChild($commandFrame);

			if ($index % 2 === 0) {
				$lineQuad = new Quad_BgsPlayerCard();
				$commandFrame->addChild($lineQuad);
				$lineQuad->setSize($width, 4);
				$lineQuad->setSubStyle($lineQuad::SUBSTYLE_BgPlayerCardBig);
				$lineQuad->setZ(-0.001);
			}

			$labelLine3 = new LabelLine($commandFrame);
			$labelLine3->addLabelEntryText($command['command'], $posX + 5, 33);
			$labelLine3->addLabelEntryText($command['explanation'], $posX + 45, 82);
			$labelLine3->render();

			$commandFrame->setY($posY);

			$posY -= 4;
			$index++;
		}

		// Render and display xml
		$this->maniaControl->getManialinkManager()->displayWidget($maniaLink, $player, 'Commands');
	}

	public function handleManialinkPageAnswer( $callback ){
		$playerLogin = $callback[1][1];

		$player = $this->maniaControl->getPlayerManager()->getPlayer( $playerLogin );

		$actionId    = $callback[1][2];
		$actionArray = explode('.', $actionId, 4);
		$action      = isset($actionArray[1]) ? $actionArray[1] : "";
		$actionClass = isset($actionArray[0]) ? $actionArray[0] : "";

		if( $action === 'SaveConfigAction' ){
			$scriptSettings = $this->maniaControl->getClient()->getModeScriptSettings();
			$currentlySetTimeLimit = $scriptSettings['S_TimeLimit'];

			if( $this->currentTimeLimit != $currentlySetTimeLimit ) {
				$this->currentTimeLimit = $currentlySetTimeLimit;
				$this->maniaControl->getChat()->sendChat("Setting the time limit via Script Settings is of temporary effect and will be overwritten by the MapTime-Plugin!");
			}
		}

		if( $actionClass === 'MapTimePlugin' && $action === 'LineClick' ) {
			if (!$this->maniaControl->getAuthenticationManager()->checkPermission($player, self::SETTING_PERMISSION_MAP_TIME_PLUGIN)) {
				$this->maniaControl->getAuthenticationManager()->sendNotAllowed($player);
				return;
			}

			$mapUID      = $actionArray[2];

			if( !isset($this->numberOfDeleteClicks[$playerLogin][$mapUID]) || $this->numberOfDeleteClicks[$playerLogin][$mapUID] !== 2 ) {
				$tempCounter = isset($this->numberOfDeleteClicks[$playerLogin][$mapUID]) ? $this->numberOfDeleteClicks[$playerLogin][$mapUID] : 0;
				$this->numberOfDeleteClicks[$playerLogin] = array();
				$this->numberOfDeleteClicks[$playerLogin][$mapUID] = $tempCounter + 1;

				$tempCounter = 0;
				return;
			}

			$this->setTimeToStandard( $mapUID );

			$this->maniaControl->getChat()->sendInformationToAdmins("Individual time limit for the clicked map deleted. Attention: This has no effect on the current time limit!");

			$this->showWidget( array(), $player );

			$this->numberOfDeleteClicks[$playerLogin] = array();

			return;
		}

		return;
	}

	private function getMapTimeLimit( $uid ) {
		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query = "SELECT timelimit FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE uid = '" . $uid . "'";
		$result = $mysqli->query($query)->fetch_assoc();

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		return isset($result['timelimit']) ? $result['timelimit'] : false;
	}

	private function getStandardTime() {
		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query  = "SELECT * FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE uid = 'standard'";
		$result = $mysqli->query($query)->fetch_assoc();

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}

		return $result['timelimit'];
	}

	private function setTimeToStandard( $uid ) {
		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query = "DELETE FROM `" . self::TABLE_MAP_PLAY_TIME . "` WHERE `uid` = '" . $uid . "'";

		$mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error, E_USER_ERROR);
		}
		return;
	}

	private function getTimeLimits() {

		//TODO: nur Maps, die auch aktuell in der Queue sind!

		$mysqli = $this->maniaControl->getDatabase()->getMysqli();
		$query  = "SELECT m.index as id, m.mxid, m.uid, m.name as name, timelimit as time FROM mc_maptime mt
					RIGHT JOIN mc_maps m ON mt.uid = m.uid
					WHERE mapName <> 'standard'
					ORDER BY mapName";

		$result = $mysqli->query($query);

		if ($mysqli->error) {
			trigger_error($mysqli->error);
			return null;
		}

		$limits = array();
		while ($record = $result->fetch_object()) {
			array_push($limits, $record);
		}

		$result->free();

		return $limits;
	}

	//TODO: Funktion zur Bereinigung der Tabelle - bereits gelöschte Maps müssen raus. Aufruf in mapBegin() oder load()
}